package coins.backend.ana;

import java.io.*;
import java.util.*;
import coins.backend.*;
import coins.backend.cfg.*;
import coins.backend.util.*;

/** Find dominance frontiers of the control flow graph. */
public class DominanceFrontiers implements LocalAnalysis {

  /** Factory class of DominanceFrontiers. */
  private static class Analyzer implements LocalAnalyzer {
    public LocalAnalysis doIt(Function func) {
      return new DominanceFrontiers
        (func.flowGraph, (Dominators)func.require(Dominators.analyzer));
    }
  }

  /** Factory singleton. */
  public static final Analyzer analyzer = new Analyzer();

  /** BasicBlk-id-indexed array, whose elements are dominance
   *   frontiers of the block. */
  public BiList frontiers[];

  private FlowGraph flowGraph;

  /** Copy of CFG timestamp to be analyzed. */
  private int timeStamp;


  private void addFrontiers(BasicBlk blk, Dominators d) {

    for (BiLink p = d.kids[blk.id].first(); !p.atEnd(); p = p.next())
      addFrontiers((BasicBlk)p.elem(), d);

    frontiers[blk.id] = new BiList();
    for (BiLink p = blk.succList().first(); !p.atEnd(); p = p.next()) {
      BasicBlk succ = (BasicBlk)p.elem();
      if (d.idom[succ.id] != blk)
        frontiers[blk.id].add(succ);
    }

    for (BiLink p = d.kids[blk.id].first(); !p.atEnd(); p = p.next()) {
      BasicBlk kid = (BasicBlk)p.elem();
      for (BiLink q = frontiers[kid.id].first(); !q.atEnd(); q = q.next()) {
        BasicBlk front = (BasicBlk)q.elem();
        if (d.idom[front.id] != blk)
          frontiers[blk.id].addNew(front);
      }
    }

  }
      

  /** Create dominance frontiers */
  public DominanceFrontiers(FlowGraph g, Dominators d) {
    flowGraph = g;
    timeStamp = flowGraph.timeStamp();

    frontiers = new BiList[g.idBound()];
    addFrontiers(g.entryBlk(), d);
  }


  /** Return true if this analysis is up to date. */
  public boolean isUpToDate() {
    return (timeStamp == flowGraph.timeStamp());
  }


  /** Dump dominance frontiers on stream out */
  public void printIt(PrintWriter out) {
    out.println();
    out.println("Dominance Frontiers:");
    for (BiLink p = flowGraph.basicBlkList.first(); !p.atEnd(); p = p.next()) {
      BasicBlk blk = (BasicBlk)p.elem();
      if (blk.dfn() != 0) {
        out.print("#" + blk.id + ": ");
        boolean top = true;
        for (BiLink f = frontiers[blk.id].first(); !f.atEnd(); f = f.next()) {
          if (!top) out.print(",");
          out.print("#" + ((BasicBlk)f.elem()).id);
          top = false;
        }
        out.println();
      }
    }
  }

}
